/**
 * Script for tree view. The tree itself is a native Firefox tree view, but with custom logic implemented.<br>
 * This single page displays all three trees - main tree, panel tree and link tree. It is achieved by changing hash in page's url - for example, to show link tree url must be changed to something like "treeview.xul#links"
 * @module  treeview
 */

var tree;
var utils;
var clipboard;

var currentView = "";

var strings;

var tabsmozInit = false;

var currentNode = null;
var hoveredNode = null;
var currentIndex = -1;
var popupShowing = false;
var nodeToggle = false;

var tempOpen = {};

var debug = false;

var OS = null;

// shorthand functions
function o(guid)
{
	return tree.getObject(guid);
}

var icons = {
	"root": "root",
	"panelroot": "folder",
	"linkroot": "linkroot",
	"panel": "panel",
	"group": "folder",
	"page": "page",
	"link": "link",
	"curation": "curation",
	"curationroot": "curationroot",
	"curationfile": "panel",
	"curationgroup": "folder",
	"curationpage": "curation"
};

/**
 * Custom nsITreeView implementation. See https://developer.mozilla.org/en-US/docs/Mozilla/Tech/XPCOM/Reference/Interface/nsITreeView for details
 * @namespace treeView
 * @type {Object}
 */
var treeView = {
	/**
	 * Array of rows in treeview.<br>
	 * format of visibleData elements: <br>
	 * {<br>
	 *   name: string<br>
	 *   guid: string<br>
	 *   container: bool<br>
	 *   level: int<br>
	 *   open: bool,<br>
	 *   properties: array of strings<br>
	 *   type: string<br>
	 * }
	 * @type {Array}
	 */
	visibleData: [],

	/**
	 * The nsITreeBoxObject interface contains information about the size and layout of a tree.
	 * @type {nsITreeBoxObject}
	 */
	treeBox: null,
	selection: null,
	rootLevel: 0,

	/**
	 * Makes object suitable for visibleData array
	 * @param  {BaseObject} node      tree node to represent in tree view
	 * @param  {int} rootLevel root level of currently displayed subtree. 0 for root, 2 for panels, 3 for groups etc.
	 * @param  {boolean} open      is node open
	 * @return {object}           see visibleData structure
	 */
	makeVisibleNode: function(node, rootLevel, open)
	{
		var icon = "chrome://store.tabsmoz/content/tree/img/" + icons[node.type] + ".png";
		if (node.type == 'link')
			icon = favicons.getFavicon(node.domain);
		if (node.type == 'link' && node.linkType == 'curation')
			icon = "chrome://store.tabsmoz/content/tree/img/curation.png";
		if (node.type == 'linkroot')
			icon = "chrome://store.tabsmoz/content/tree/img/linkroot.png";
		if (node.type == 'page')
		{
			var link = o(node.link);
			if (link)
				icon = favicons.getFavicon(link.domain);
		}
		return {
			name: node.name,
			guid: node.guid,
			container: (node.children.length == 0 || node.type == 'page') ? false : true,
			open: node.open,
			level: node.level - rootLevel,
			type: node.type,
			icon: icon,
			isRoot: node.level == rootLevel,
			properties: []
		};
	},

	/**
	 * populate visibleData from main tree
	 * @param  {BaseObject} root root node of subtree
	 * @param {boolean} open is tree opened
	 */
	prepareTree: function(root, open)
	{
		if(!this.treeBox)
			return;
		var data = [];
		var rootLevel = root.level;
		var self = this;

		function traverse(node)
		{
			data.push(self.makeVisibleNode(node, rootLevel, node.open));
			debug && console.log("[TV] " + node.name + ": " + node.open);
			if (data[data.length - 1].open && node.type != 'page')
				for (var i = 0; i < node.children.length; i++)
					traverse(o(node.children[i]));
		}
		traverse(root);

		this.rootLevel = rootLevel;		
		var l = this.visibleData.length;
		this.visibleData = data;
		this.treeBox.rowCountChanged(0, data.length - l);
		this.treeBox.invalidate();
		//console.log("[TV] Preparing tree ", data.length - l);
		debug && console.log('[TV] ', this.visibleData.length);
		setTimeout(function()
		{
			debug && console.log("[TV] Updating selection");
			treeView.updateSelection();
		}, 300);
	},

	/**
	 * Adds new node to tree view according to its position
	 * @param  {BaseObject} node node to represent in tree view
	 * @param  {object} data additional data about node
	 * @param {int} data.position position among node's siblings
	 */
	insertNode: function(node, data)
	{
		if (node.type == 'subpage' || node.type == 'curationlink' || node.type == 'rsslink')
			return;
		var parent = null,
			idx = 0;
		for (var i = 0; i < this.visibleData.length; i++)
		{
			if (this.visibleData[i].guid == node.parent)
			{
				parent = o(node.parent);
				idx = i;
				break;
			}
		}
		if (!parent)
			return;

		if (this.visibleData[idx].container && this.visibleData[idx].open)
		{
			var newNode = this.makeVisibleNode(node, this.rootLevel, node.open);
			var i = idx + 1;
			var pos = data.position;
			var level = [];
			while (i < this.visibleData.length && this.visibleData[i].level > this.visibleData[idx].level)
			{
				if (this.visibleData[i].level == this.visibleData[idx].level + 1)
					level.push(i);
				i++;
			}
			level.push(i);
			pos = level[pos];
			debug && console.log("[TV] Inserting: position: " + data.position + "; Index of replace: " + pos + "; parent idx: " + idx, level);
			this.visibleData.splice(pos, 0, newNode);
			debug && console.log("[TV] New node: ", node, newNode);
			this.treeBox.rowCountChanged(pos, 1);
		}
		if (!this.visibleData[idx].container)
		{
			this.visibleData[idx].container = true;
			this.visibleData[idx].open = false;
			this.treeBox.invalidateRow(idx);
		}
		this.updateSelection();
	},

	/**
	 * Removes node from tree view and its children
	 * @param  {BaseObject} node node to remove
	 */
	removeNode: function(node)
	{
		// important check: user can copy node, then delete it, so we should clear the clipboard
		if (node == clipboard.node)
		{
			clipboard.clear();
		}
		//console.log("[TV] Deleting node " + node.name);
		var data = this.visibleData;
		for (var i = 0; i < this.visibleData.length; i++)
		{
			if (data[i].guid == node.guid)
			{
				var parent = tree.getObject(node.parent);
				debug && console.log("[TV] Parent children ", node.type, node.name);
				if (parent && parent.children.length == 0)
				{
					debug && console.log("[TV] This was the last child")
					for (var j = i - 1; j >= 0; j--)
					{
						if (this.visibleData[j].guid == parent.guid)
							this.visibleData[j].container = false;
						this.treeBox.invalidateRow(j);
						break;
					}
				}
				this.visibleData.splice(i, 1);
				this.treeBox.rowCountChanged(i + 1, -1);
				/*for(var j = i + 1; j < data.length && data[j].level > node.level; j++)
				data.splice(i, j - i);
				this.treeBox.rowCountChanged(j + 1, j - i);*/
				break;
			}
		}
		this.updateSelection();
	},

	/**
	 * Updates node info (e.g. name, children position)
	 * @param  {BaseObject} node node to update
	 * @param  {object} data key-value pairs of information about update
	 * @param {string[]} data.properties array of property names that need to be updated
	 */
	updateNode: function(node, data)
	{
		var idx = -1;
		for (var i = 0; i < this.visibleData.length; i++)
		{
			if (this.visibleData[i].guid == node.guid)
			{
				idx = i;
				break;
			}
		}
		if (idx == -1)
			return;
		
		//console.log("[TV] Updating node " + idx);
		//console.log("[TV] Updating node " + this.visibleData[idx].name);

		if (node.type !== 'link' && node.type !== 'linkroot' && node.type !== 'panelroot' && node.type !== 'root')
		{
			if (data.properties.indexOf('lastChild') > -1)
				this.updateSelection(node.lastChild);
		}

		if (node.type == 'page')
		{	
			var link = o(node.link);
			if(link)
				this.visibleData[idx].icon = favicons.getFavicon(link.domain);
		}

		if (node.type == 'link')
		{
			this.visibleData[idx].icon = favicons.getFavicon(node.domain);
		}

		// rename node
		if (data.properties.indexOf('name') > -1)
		{
			this.visibleData[idx].name = node.name;
			this.sortLinkTree();
		}

		this.treeBox.invalidateRow(idx);

		// swap children of a node (only pages for now)
		if (data.properties.indexOf('swap') > -1)
		{
			var i1 = idx;
			while (i1 < this.visibleData.length && this.visibleData[i1].guid != data.data.c1) i1++;
			var i2 = idx;
			while (i2 < this.visibleData.length && this.visibleData[i2].guid != data.data.c2) i2++;
			if (i1 != i2 && i1 != idx && i2 != idx)
			{
				var l1 = i1 + 1;
				while (l1 < this.visibleData.length && this.visibleData[l1].level > this.visibleData[i1].level) l1++;
				var l2 = i2 + 1;
				while (l2 < this.visibleData.length && this.visibleData[l2].level > this.visibleData[i2].level) l2++;
				var n1 = this.visibleData.splice(i1, l1 - i1);
				var s = i2; 
				if(i1<i2) 
					s = l2 - l1 + i1;
				Array.prototype.splice.apply(this.visibleData, [s, 0].concat(n1));
				/*for(var i=0; i<this.visibleData.length; i++) {
					this.treeBox.invalidateRow(i);
				}*/
				this.treeBox.invalidate();
			}
		}

	},

	updateSelection: function(guid)
	{
		var sel = -1;
		if(!guid && currentView == '#panels')
		{
			var page = tree.getSelected('page');
			if(!page)
				return;
			guid = page.guid;
		}
		var child = guid;
		for (var i = 0; i < this.visibleData.length; i++)
		{
			if (this.visibleData[i].guid == child)
			{
				sel = i;
				break;
			}
		}
		//this.highlightNode(sel, true);
		if (sel > -1)
			this.selection.select(sel);
	},

	/**
	 * Sorts link tree alphabetically
	 */
	sortLinkTree: function()
	{
		//return;

		debug && console.log('[TV] Sorting');
		var rootIdx = -1;
		for (var i = 0; i < this.visibleData.length; i++)
		{
			if (o(this.visibleData[i]).type == 'linkroot')
			{
				rootIdx = i;
				break;
			}
		}
		if (rootIdx < 0)
			return;

		debug && console.log('[TV] Root index: ' + rootIdx);

		function compareLinks(a, b)
		{
			return o(a.guid).name.localeCompare(o(b.guid).name);
		}

		var data = this.visibleData;
		var newLinkTree = [];

		function buildTree(index)
		{
			var row = data[index];
			var node = {
				index: index,
				guid: row.guid,
				children: []
			};
			var i = index + 1;
			while (i < data.length && data[i].level > row.level)
			{
				if (data[i].level == row.level + 1)
					node.children.push(buildTree(i));
				i++;
			}
			return node;
		}

		var lroot = buildTree(rootIdx);

		debug && console.log('[TV]', lroot);

		function traverse(node)
		{
			if (node.children.length)
				node.children.sort(compareLinks);
			newLinkTree.push(data[node.index]);
			for (var i = 0; i < node.children.length; i++)
				traverse(node.children[i]);
		}

		traverse(lroot);

		for (var i = 0; i < newLinkTree.length; i++)
		{
			debug && console.log("[TVS] " + i + ": (" + newLinkTree[i].level + ") " + o(newLinkTree[i].guid).name);
			//newLinkTree[i].data = data[newLinkTree[i].start];
		}
		//data.splice(rootIdx, newLinkTree.length);
		for (var i = 0; i < newLinkTree.length; i++)
		{
			//data.push(newLinkTree[i]);
			data[rootIdx + i] = newLinkTree[i];
		}
	},

	grayNode: function(idx)
	{
		this.visibleData[idx].properties.push("grayed");
		this.treeBox.invalidateRow(idx);
	},

	/**
	 * Adds 'hover' effect on given row in tree view. Used to highlight node when something is dragged over
	 * @param  {int} idx index of row
	 */
	highlightNode: function(idx, hard)
	{
		for (var i = 0; i < this.visibleData.length; i++)
		{
			this.visibleData[i].properties = [];
			this.treeBox.invalidateRow(i);
		}
		if (idx >= 0 && idx < this.visibleData.length)
		{
			if(OS == 'Darwin') 
				hard = true;
			this.visibleData[idx].properties = hard ? ['selected'] : ['hover'];
			this.treeBox.invalidateRow(idx);
		}
	},

	updateIcon: function(domain)
	{
		var icon = favicons.getFavicon(domain);
		for (var i = 0; i < this.visibleData.length; i++)
		{
			if (this.visibleData[i].type == 'link' && o(this.visibleData[i].guid).domain == domain)
			{
				this.visibleData[i].icon = icon;
				this.treeBox.invalidateRow(i);
			}
			if (this.visibleData[i].type == 'page')
			{
				var link = o( o(this.visibleData[i].guid).link );
				if (link && link.domain == domain)
				{
					this.visibleData[i].icon = icon;
					this.treeBox.invalidateRow(i);
				}
			}
		}
	},

	get rowCount()
	{
		return this.visibleData.length;
	},
	setTree: function(treeBox)
	{
		this.treeBox = treeBox;
	},
	getCellText: function(idx, column)
	{
		return this.visibleData[idx].name;
	},
	isContainer: function(idx)
	{
		return this.visibleData[idx].container;
	},
	isContainerOpen: function(idx)
	{
		return this.visibleData[idx].open;
	},
	isContainerEmpty: function(idx)
	{
		return false;
	},
	isSeparator: function(idx)
	{
		return false;
	},
	isSorted: function()
	{
		return false;
	},
	isEditable: function(idx, column)
	{
		return false;
	},

	getParentIndex: function(idx)
	{
		if (idx < 0 || idx >= this.visibleData.length) return -1;
		if (this.visibleData[idx].isRoot) return -1;
		for (var t = idx - 1; t >= 0; t--)
		{
			if (this.visibleData[t].level == this.visibleData[idx].level - 1) return t;
		}
		return 0;
	},
	getLevel: function(idx)
	{
		return this.visibleData[idx].level;
	},
	hasNextSibling: function(idx, after)
	{
		return false;
		var thisLevel = this.getLevel(idx);
		for (var t = idx + 1; t < this.visibleData.length; t++)
		{
			var nextLevel = this.getLevel(t);
			if (nextLevel == thisLevel) return true;
			if (nextLevel < thisLevel) break;
		}
		return false;
	},
	toggleOpenState: function(idx, open)
	{
		debug && console.log("[TV] Open/Close!");
		nodeToggle = true;
		var item = this.visibleData[idx];

		if (!item.container) return;

		if (hoveredNode && !open)
			return;

		var node = o(this.visibleData[idx].guid);
		if (item.open && !open)
		{
			item.open = false;
			node.open = false;
			var thisLevel = this.getLevel(idx);
			var deletecount = 0;
			for (var t = idx + 1; t < this.visibleData.length; t++)
			{
				if (this.getLevel(t) > thisLevel) deletecount++;
				else break;
			}
			if (deletecount)
			{
				this.visibleData.splice(idx + 1, deletecount);
				this.treeBox.rowCountChanged(idx + 1, -deletecount);
			}
		}
		else
		{
			item.open = true;
			node.open = true;
			var rowsAdded = 0;
			var self = this;

			function traverse(node)
			{
				for (var i = 0; i < node.children.length; i++)
				{
					var c = o(node.children[i]);
					rowsAdded++;
					self.visibleData.splice(idx + rowsAdded, 0, self.makeVisibleNode(c, self.rootLevel, c.open));
					if (c.open)
						traverse(c);
				}
			}
			traverse(node);
			this.treeBox.rowCountChanged(idx + 1, rowsAdded);
		}
		this.treeBox.invalidateRow(idx);

		tree.toggleOpen(item.guid, item.open);
	},

	drop: function(idx, orientation, dataTransfer)
	{
		function copymove(src, dst)
		{
			src = tree.getObject(src);
			if(!clipboard.canManipulate([src]))
				return;
			dst = dst || node;
			if (dataTransfer.dropEffect == 'move')
			{ 
				tree.moveSubtree(src, dst);
			}
			if (dataTransfer.dropEffect == 'copy')
			{
				tree.copySubtree(src, dst);
			}
		}

		hoveredNode = null;
		if(idx < 0 || idx >= this.visibleData.length)
			return;
		var node = tree.getObject(this.visibleData[idx].guid);
		var data = utils.extractURL(dataTransfer);
		debug && console.log("[TREE] " + node.name + "; " + orientation, data);
		// copy rss link into curation folder. Very special case
		if(data.nodelist && data.nodetype == 'rsslink' && node.linkType == 'curation')
		{
			for(var i=0; i<data.nodelist.length; i++)
			{
				var r = tree.getObject(data.nodelist[i]);
				var c = tree.createCurationLink({
					parent: node.guid,
					url: r.url,
					headline: r.title,
					author: r.author,
					description: r.content,
					pubDate: r.pubDate,
					imageUrl: r.img,
					tags: r.tags
				});
			}
			return;
		}
		if (data.nodelist && clipboard.compatible(node.type, node, data.nodetype))
		{
			var nodes = data.nodelist;
			for(var i=0; i<nodes.length; i++)
			{
				copymove(nodes[i], node);
			}
			this.highlightNode(-1);
			return;
		}
		if (node.type == "link" || node.type == 'linkroot')
		{
			if(data.url && !data.link)
			{
				if(node.linkType == 'curation')
				{
					var link = tree.createCurationLink(
					{
						name: data.name,
						url: data.url,
						parent: node.guid
					}, true);
				} else {
					var link = tree.createLink(
					{
						name: data.name,
						url: data.url,
						parent: node.guid
					}, true);
				}
				window.parent.Tabsmoz.showProperties(link, true,
				{
					createdByDragDrop: data.url
				});
			}
		} else if (data.web) {
			switch(node.type) {
				case "panel": // create group
					var group;
					group = tree.createGroup(
					{
						name: "",
						link: data.link, //
						url: data.url,
						parent: node.guid
					}, true);
					window.parent.Tabsmoz.showProperties(group, true,
					{
						pagename: data.name,
						url: data.url,
						link: data.link
					});
				break;

				case "group": // create page
					var showProperties = utils.getSetting('showPropertiesOnCreate', 'bool');
					var page = tree.createPage(
					{
						parent: node.guid,
						link: data.link,
						url: data.url,
						name: data.name,
						select: true
					}, showProperties);
					if (showProperties)
						window.parent.Tabsmoz.showProperties(page, true,
						{
							createdByDragDrop: data.url
						});
				break;

				case "page": // create subpage
					var showProperties = utils.getSetting('showPropertiesOnCreate', 'bool');
					var page = tree.createPage(
					{
						parent: node.parent,
						link: data.link,
						name: data.name,
						select: true
					}, showProperties);
					if (showProperties)
						window.parent.Tabsmoz.showProperties(page, true,
						{
							createdByDragDrop: data.url
						});
				break;

				case "curationgroup":
					var link = tree.getObject(data.link);
					if(link && link.linkType == 'curation')
					{
						var page = tree.createCurationPage(
						{
							parent: node.guid,
							link: link.guid,
							name: data.name
						}, true);
						window.parent.Tabsmoz.showProperties(page, true,
						{
							createdByDragDrop: true
						});
					}
				break;
			}
		}
		if (data.node || data.link)
		{
			var src = tree.getObject(data.node || data.link	);
			if (clipboard.compatible.call(
			{
				type: src.type
			}, node.type))
			{
				copymove(src, node);
			}
		}
		this.highlightNode(-1);
	},

	getImageSrc: function(idx, column)
	{
		return this.visibleData[idx].icon;
	},
	getProgressMode: function(idx, column) {},
	getCellValue: function(idx, column) {},
	cycleHeader: function(col, elem) {},
	selectionChanged: function() {},
	cycleCell: function(idx, column) {},
	performAction: function(action) {},
	performActionOnCell: function(action, index, column) {},
	getRowProperties: function(idx)
	{
		if (this.visibleData[idx].properties)
			return this.visibleData[idx].properties.join(' ');
		return "";
	},
	getCellProperties: function(idx, column, prop)
	{
		/*if(this.visibleData[idx].properties)
      return this.visibleData[idx].properties.join(' ');*/
		return "";
	},
	getColumnProperties: function(column, element, prop)
	{
		return "";
	},
};

/**
 * Handles 'update' event on a tree
 * @param  {object} data key-value pairs of event data
 * @param {BaseObject} data.target modified node of a tree
 * @param {object} data.data additional info
 */
function updateTree(data)
{
	switch (data.type)
	{
		case 'add':
			treeView.insertNode(data.target, data.data);
			break;

		case 'delete':
			treeView.removeNode(data.target);
			break;

		case 'update':
			treeView.updateNode(data.target, data);
			if(currentView == '#page')
			{
				setPageTree();
			}
			break;
	}
}

/**
 * Called before context menu is shown. Controls what menu entries should be active, and sets proper labels
 * @param  {Event} event
 */
function showTreePopup(event)
{
	function isEditable(type)
	{
		return type !== 'root' && type !== 'linkroot' && type !== 'panelroot' && type !== 'curationroot';
	}
	var idx = currentIndex; //document.getElementById('mainTree').currentIndex;
	//debug && console.log("[TV] ", document.getElementById('mainTree').currentItem);
	if (idx < 0)
	{
		debug && console.log("[TV] NO CONTEXT");
		$('tabsmoz-tree-menu-cut').disabled = true;
		$('tabsmoz-tree-menu-copy').disabled = true;
		$('tabsmoz-tree-menu-paste').disabled = true;
		$('tabsmoz-tree-menu-properties').disabled = true;
		$('tabsmoz-tree-menu-delete').disabled = true;
		$('tabsmoz-tree-menu-new').hidden = true;
		$('tabsmoz-tree-menu-newsep').hidden = true;
		$('tabsmoz-tree-menu-new-link').hidden = true;
		return;
	}
	else
	{
		var node = o(treeView.visibleData[idx].guid);
	}

	$('tabsmoz-tree-export-curation').hidden = node.type !== 'curationfile';

	function $(id)
	{
		return document.getElementById(id);
	}
	$('tabsmoz-tree-menu-new-link').label = strings ? strings.getString('new.link') : "New Link Folder";
	var multiSelect = false;
	currentNode = node;
	var disableCopyPaste = !clipboard.canManipulate([node]);
	var disableDelete = !clipboard.canDelete([node]);
	$('tabsmoz-tree-menu-cut').disabled = disableCopyPaste;
	$('tabsmoz-tree-menu-copy').disabled = disableCopyPaste;
	$('tabsmoz-tree-menu-paste').disabled = multiSelect || !clipboard.compatible(node.type, node);
	$('tabsmoz-tree-menu-properties').disabled = multiSelect || node.type == 'rsslink';
	$('tabsmoz-tree-menu-delete').disabled = disableDelete;

	var t = node.type;
	if(t == 'linkroot')
	{
		$('tabsmoz-tree-menu-delete').disabled = true;
		t = 'link';
	}
	if(t == 'link')
	{
		$('tabsmoz-tree-menu-new').hidden = true;
		$('tabsmoz-tree-menu-new-link').hidden = false;
	} else {
		$('tabsmoz-tree-menu-new').hidden = false;
		$('tabsmoz-tree-menu-new-link').hidden = true;
	}
	if(t == 'rsslink')
	{
		$('tabsmoz-tree-menu-new').disabled = true;
		$('tabsmoz-tree-menu-delete').disabled = true;
	}
	if (t !== 'page' && t !== 'curationpage')
	{
		$('tabsmoz-tree-menu-newsep').hidden = false;
		$('tabsmoz-tree-menu-new').label = strings.getString('new.' + t);
	} else {
		$('tabsmoz-tree-menu-new').hidden = true;
		$('tabsmoz-tree-menu-newsep').hidden = true;
	}
	$('tabsmoz-tree-menu-export').hidden = t != 'panelroot';
	$('tabsmoz-tree-menu-import').hidden = t != 'panelroot';

	$('tabsmoz-tree-menu-delete').label = strings.getString('delete.' + t);
	$('tabsmoz-tree-menu-cut').label = strings.getString('cut.' + t);
	$('tabsmoz-tree-menu-copy').label = strings.getString('copy.' + t);
	$('tabsmoz-tree-menu-paste').label = strings.getString('paste.' + (clipboard.node == null ? "null" : clipboard.node.type));

	selectedNode = node;
	//console.log("[LV] Selected node", selectedNode);
}

function hideTreePopup(event)
{
	popupShowing = false;
	debug && console.log('[TV] Hide popup');
}

/**
 * Just shows properties dialog for a node
 * @param  {Event} event
 */
function showProperties(event)
{
	window.parent.Tabsmoz.showProperties(currentNode);
}

/**
 * Function to create new node. Called from context menu
 * @todo refactoring: выделить в отдельный метод в tree, например
 * @param  {Event} event
 */
function newNode(event, linkType)
{
	switch (currentNode.type)
	{
		case 'panelroot':
			var panel = tree.createPanel(
			{
				name: " ",
			}, true);
			window.parent.Tabsmoz.showProperties(panel, true);
			break;
		case 'panel':
			var group = tree.createGroup(
			{
				name: " ",
				parent: currentNode.guid
			}, true);
			window.parent.Tabsmoz.showProperties(group, true);
			break;
		case 'group':
			var page = tree.createPage(
			{
				name: " ",
				parent: currentNode.guid
			}, true);
			window.parent.Tabsmoz.showProperties(page, true);
			break;
		case 'linkroot':
		case 'link':
			linkType = linkType || 'folder';
			//console.log("[TV] LT " + linkType);
			var link = tree.createLink(
			{
				name: "",
				url: " ",
				parent: currentNode.guid,
				linkType: linkType
			}, true);
			window.parent.Tabsmoz.showProperties(link, true);
			break;
		case 'curationroot':
			var file = tree.createCurationFile({}, true);
			window.parent.Tabsmoz.showProperties(file, true);
		break;
		case 'curationfile':
			var group = tree.createCurationGroup({ parent: currentNode.guid }, true);
			window.parent.Tabsmoz.showProperties(group, true);
		break;
		case 'curationgroup':
			var page = tree.createCurationPage({ parent: currentNode.guid }, true);
			window.parent.Tabsmoz.showProperties(page, true);
		break;
	}
}

function exportCuration(event)
{
	if(currentNode.type == 'curationfile')
	{
		window.parent.Tabsmoz.exportCurationFile(currentNode);
	}
}

/**
 * Function to remove node. Called from context menu
 * @param  {Event} event
 */
function removeNode(event)
{
	tree.removeNode(currentNode);
}

/**
 * Inits clipboard with current node and 'cut' action. Called from context menu
 * @param  {Event} event
 */
function cutNode(event)
{
	clipboard.set(currentNode, 'cut');
	//var idx = document.getElementById('mainTree').currentIndex;
	treeView.grayNode(currentIndex);
}

/**
 * Inits clipboard with current node and 'copy' action. Called from context menu
 * @param  {Event} event
 */
function copyNode(event)
{
	clipboard.set(currentNode, 'copy');
}

/**
 * Performing 'paste' operation. Called from context menu
 * @param  {Event} event
 */
function pasteNode(event)
{
	var data = clipboard.data;
	if (clipboard.action == 'cut')
	{
		for(var i=0; i<data.length; i++)
			tree.moveSubtree(data[i], currentNode);
		clipboard.clear();
	}
	if (clipboard.action == 'copy')
	{
		for(var i=0; i<data.length; i++)
			tree.copySubtree(data[i], currentNode);
	}
}

function exportPanelTree()
{
	window.parent.Tabsmoz.exportPanelTree();
}

function importPanelTree()
{
	window.parent.Tabsmoz.importPanelTree();
}

/**
 * Handles node selection
 * @param  {Event} event
 */
function treeSelect(event)
{
	debug && console.log("[TV] Select!");
}

function treeMouseUp(event)
{
	if (event.button != 0)
		return;
	var row = {},
		col = {},
		child = {};
	treeView.treeBox.getCellAt(event.clientX, event.clientY, row, col, child);
	var idx = row.value;
	debug && console.log("[TV] Mouse Up! ", row, col);
	if (idx > -1 && idx < treeView.visibleData.length)
	{
		popupShowing = false;
		setTimeout(function()
		{
			if (!nodeToggle)
			{
				if(window.parent.Tabsmoz.getManagementMode() == 'local')
				{
					tree.openLinkInPage(tree.getSelected('page'), treeView.visibleData[idx].guid);
				} else {
					window.parent.Tabsmoz.selectNode(treeView.visibleData[idx].guid);
				}
			}
			nodeToggle = false;
		}, 100);
	}
}

/**
 * Handles start of dragging node, sets dataTransfer object with data according to node properties
 * @param  {Event} event
 */
function treeDragStart(event)
{
	debug && console.log("[TV] Started dragging!");
	var idx = document.getElementById('mainTree').currentIndex;
	var node = tree.getObject(treeView.visibleData[idx].guid);
	if (node.type == 'link' || node.type == 'linkroot')
	{
		event.dataTransfer.setData('text/x-moz-url-data', node.url);
		event.dataTransfer.setData('text/x-moz-url-desc', node.name);
		event.dataTransfer.setData("text/uri-list", node.url);
		event.dataTransfer.setData('text/x-tabsmoz-link', node.guid);
		event.dataTransfer.setData('text/x-tabsmoz-linktype', node.linkType);
		debug && console.log("[TV] Dragging " + node.name);
	}
	if (node.type == "page" || node.type == "group" || node.type == "panel")
	{
		if (node.type !== "panel")
			event.dataTransfer.setData('text/x-tabsmoz-tab', node.guid);
		event.dataTransfer.setData('text/x-tabsmoz-node', node.guid);
		debug && console.log("[TV] Dragging " + node.name);
	}
	if (node.type == "curationpage" || node.type == "curationgroup")
	{
		event.dataTransfer.setData('text/x-tabsmoz-node', node.guid);
	}
}

/**
 * Handles dragging something over node. Opens closed ones and highlights node under cursor
 * @param  {Event} event drag event, used to get node under cursor
 */
function treeDragOver(event)
{
	var row = {},
		col = {},
		child = {};
	treeView.treeBox.getCellAt(event.clientX, event.clientY, row, col, child);
	debug && console.log("[TV] Drag " + row.value);
	if (row.value > -1 && row.value < treeView.visibleData.length)
	{
		var node = treeView.visibleData[row.value];
		treeView.highlightNode(row.value);
		var data = utils.extractURL(event.dataTransfer);
		var over = tree.getObject(data.node);
		var under = tree.getObject(node.guid);
		if(data.node && over.type == under.type && over.type !== 'link' && over.type !== 'linkroot')
		{
			tree.swapNodes(over, under);
		}
		if (node.container && !node.open && hoveredNode !== row.value)
		{
			hoveredNode = row.value;
			var timeout = utils.getSetting("hoverLinkDelay", "int");
			setTimeout(function()
			{
				if (row.value == hoveredNode && !node.open)
				{
					debug && console.log("[TREE] Toggling node, fuck");
					treeView.toggleOpenState(row.value, true);
				}
			}, timeout);
		}
	}
	else
		treeView.highlightNode(-1);
	event.preventDefault();
}

function treeBlur(event)
{
	hoveredNode = null;
	treeView.highlightNode(-1);
}

function setPageTree() 
{
	var page = tree.getSelected('page');
	var link = tree.getObject(page.link) || tree.getLinkRoot();
	//console.log("[TV] setPageTree");
	treeView.prepareTree(link);
}

/**
 * Handles changes of tree type (main tree, panel tree or link tree)
 */
function detectHash()
{
	//console.log("HASH: " + location.hash);
	if (currentView !== location.hash)
	{
		document.getElementById("mainTree").view = treeView;
		currentView = location.hash;
		switch (currentView)
		{
			case '#main':
				treeView.prepareTree(tree.getRoot());
				break;

			case '#panels':
				treeView.prepareTree(tree.getPanelRoot());
				break;

			case '#links':
				treeView.prepareTree(tree.getLinkRoot());
				break;

			case '#page':
				setPageTree();
				break;

			case '#curation':
				treeView.prepareTree(tree.getCurationRoot());
				break;
		}
		clipboard.clear();
	}
}

var refreshFlag = true;
function refreshTree()
{
	//console.log("[TV] Refreshing");
	function realRefresh()
	{
		currentView = "";
		detectHash();
		refreshFlag = true;
	}
	if(refreshFlag)
	{
		refreshFlag = false;
		setTimeout(function() {
			realRefresh();
		}, 400);
	}
}

function refreshSelection(guid)
{
	treeView.updateSelection(guid);
}

function updateFavicon(data)
{
	treeView.updateIcon(data.domain);
}

function showTree(data, tree, selected)
{};

/**
 * Inits tree view. Called on page load
 */
function init()
{
	tree = window.parent.Tabsmoz.getTree();
	utils = window.parent.Tabsmoz.getUtils();
	favicons = window.parent.Tabsmoz.getFavicons();
	clipboard = window.parent.Tabsmoz.clipboard;

	OS = utils.getOSName();

	document.getElementById("mainTree").view = treeView;
	window.addEventListener("hashchange", detectHash, false);

	strings = document.getElementById('treeview-strings');	
	// disable selection on right click. Still needs improvement
	document.getElementById("mainTree").addEventListener("mousedown", function(event)
	{
		debug && console.log("[TV] MouseDown!");
		if (event.button != 2)
		{
			return;
		}
		popupShowing = true;
		var row = {},
			col = {},
			child = {};
		treeView.treeBox.getCellAt(event.clientX, event.clientY, row, col, child);
		currentIndex = row.value;
		debug && console.log("[TV] Right click " + row.value);
		if (row.value > -1 && row.value < treeView.visibleData.length)
		{
			var node = treeView.visibleData[row.value];
			debug && console.log("[TV] Node: " + node.name);
		}

		/*event.stopPropagation();
    event.preventDefault();*/
	}, true);
	console.log("[TV] Init")
	detectHash();
	tabsmozInit = true;
}

function waitParent()
{
	if(window.parent.Tabsmoz && window.parent.Tabsmoz.getTree)
		init();
	else
		setTimeout(function() {
			waitParent();
		}, 1000);
}